@@ -79,6 +79,11 @@ struct _Imagedisplay {
 	double scale;
 	double x, y;
 
+	/* The size of physical display pixels in gtk coordinates, eg. for a 200%
+	 * desktop this would be 0.5.
+	 */
+	double pixel_size;
+
 	/* Draw the screen in debug mode.
 	 */
 	gboolean debug;
@@ -126,6 +131,9 @@ enum {
 	 */
 	PROP_DEBUG,
 
+	/* Read out display density with this.
+	 */
+	PROP_PIXEL_SIZE,
 };
 
 enum {
@@ -353,9 +361,8 @@ imagedisplay_tilecache_changed(Tilecache *tilecache,
 	Imagedisplay *imagedisplay)
 {
 #ifdef DEBUG
-	printf("imagedisplay_tilecache_changed: %d x %d\n",
-		tilecache->tilesource->image_width,
-		tilecache->tilesource->image_height);
+	printf("imagedisplay_tilecache_changed: imagedisplay = %p\n",
+		imagedisplay);
 #endif /*DEBUG*/
 
 	imagedisplay->image_rect.width = tilecache->tilesource->image_width;
@@ -477,6 +484,8 @@ imagedisplay_set_property(GObject *object,
 {
 	Imagedisplay *imagedisplay = (Imagedisplay *) object;
 
+	double d;
+
 #ifdef DEBUG
 	{
 		g_autofree char *str = g_strdup_value_contents(value);
@@ -527,8 +536,7 @@ imagedisplay_set_property(GObject *object,
 		break;
 
 	case PROP_TILESOURCE:
-		imagedisplay_set_tilesource(imagedisplay,
-			g_value_get_object(value));
+		imagedisplay_set_tilesource(imagedisplay, g_value_get_object(value));
 		break;
 
 	case PROP_BESTFIT:
@@ -575,6 +583,15 @@ imagedisplay_set_property(GObject *object,
 		gtk_widget_queue_draw(GTK_WIDGET(imagedisplay));
 		break;
 
+	case PROP_PIXEL_SIZE:
+		d = g_value_get_double(value);
+		if (imagedisplay->pixel_size != d) {
+			imagedisplay->pixel_size = d;
+			imagedisplay_layout(imagedisplay);
+			gtk_widget_queue_draw(GTK_WIDGET(imagedisplay));
+		}
+		break;
+
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
 		break;
@@ -634,6 +651,10 @@ imagedisplay_get_property(GObject *object,
 		g_value_set_boolean(value, imagedisplay->debug);
 		break;
 
+	case PROP_PIXEL_SIZE:
+		g_value_set_double(value, imagedisplay->pixel_size);
+		break;
+
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
 		break;
@@ -658,42 +679,43 @@ imagedisplay_snapshot(GtkWidget *widget, GtkSnapshot *snapshot)
 		&GRAPHENE_RECT_INIT(0, 0,
 			gtk_widget_get_width(widget), gtk_widget_get_height(widget)));
 
+	gtk_snapshot_save(snapshot);
+
+	/* This can change on each repaint as windows are dragged.
+	 */
+	GtkNative *native = gtk_widget_get_native(widget);
+	GdkSurface *surface = gtk_native_get_surface(native);
+	double pixel_size = 1.0 / gdk_surface_get_scale(surface);
+	g_object_set(imagedisplay, "pixel-size", pixel_size, NULL);
+
+#ifdef HAVE_GTK_SNAPSHOT_SET_SNAP
+	gtk_snapshot_set_snap(snapshot, GSK_RECT_SNAP_ROUND);
+#endif /*HAVE_GTK_SNAPSHOT_SET_SNAP*/
+
+	gtk_snapshot_scale(snapshot, pixel_size, pixel_size);
+
 	graphene_rect_t paint;
-	paint.origin.x = imagedisplay->paint_rect.left;
-	paint.origin.y = imagedisplay->paint_rect.top;
-	paint.size.width = imagedisplay->paint_rect.width;
-	paint.size.height = imagedisplay->paint_rect.height;
+	paint.origin.x = imagedisplay->paint_rect.left / pixel_size;
+	paint.origin.y = imagedisplay->paint_rect.top / pixel_size;
+	paint.size.width = imagedisplay->paint_rect.width / pixel_size;
+	paint.size.height = imagedisplay->paint_rect.height / pixel_size;
 
 	if (imagedisplay->tilecache &&
-		imagedisplay->tilecache->tiles)
+		imagedisplay->tilecache->n_levels > 0)
 		tilecache_snapshot(imagedisplay->tilecache, snapshot,
-			imagedisplay->scale, imagedisplay->x, imagedisplay->y,
+			imagedisplay->scale / pixel_size, 
+			imagedisplay->x / pixel_size, 
+			imagedisplay->y / pixel_size,
 			&paint, imagedisplay->debug);
 
-	// draw any overlays
+	// undo snap and scale
+	gtk_snapshot_restore(snapshot);
+
+	// draw any overlays back in the regular coordinate space
 	imagedisplay_overlay_snapshot(imagedisplay, snapshot);
 
+	// end of clip
 	gtk_snapshot_pop(snapshot);
-
-	/* It's unclear how to do this :( maybe we're supposed to get the base
-	 * widget class to do it? Draw it ourselves for now.
-	if (gtk_widget_has_focus(widget)) {
-		GskRoundedRect outline;
-
-		gsk_rounded_rect_init_from_rect(&outline,
-			&GRAPHENE_RECT_INIT(
-				3,
-				3,
-				gtk_widget_get_width(widget) - 6,
-				gtk_widget_get_height(widget) - 6),
-			5);
-
-		gtk_snapshot_append_border(snapshot,
-			&outline,
-			(float[4]){ 2, 2, 2, 2 },
-			(GdkRGBA[4]){ BORDER, BORDER, BORDER, BORDER });
-	}
-	 */
 }
 
 static void
@@ -838,6 +860,13 @@ imagedisplay_class_init(ImagedisplayClass *class)
 			FALSE,
 			G_PARAM_READWRITE));
 
+	g_object_class_install_property(gobject_class, PROP_PIXEL_SIZE,
+		g_param_spec_double("pixel-size",
+			_("Pixel size"),
+			_("Size of hardware display pixels in gtk coordinates"),
+			0.0, 10.0, 0.0,
+			G_PARAM_READWRITE));
+
 	g_object_class_override_property(gobject_class,
 		PROP_HADJUSTMENT, "hadjustment");
 	g_object_class_override_property(gobject_class,
@@ -863,6 +892,8 @@ imagedisplay_class_init(ImagedisplayClass *class)
 		g_cclosure_marshal_VOID__OBJECT,
 		G_TYPE_NONE, 1,
 		GTK_TYPE_SNAPSHOT);
+
+	gtk_widget_class_set_css_name(widget_class, "imagedisplay");
 }
 
 Imagedisplay *
@@ -882,7 +913,7 @@ imagedisplay_new(Tilesource *tilesource)
 }
 
 /* image	level0 image coordinates ... this is the coordinate space we
- *		pass down to tilecache
+ *			pass down to tilecache
  *
  * gtk		screen cods, so the coordinates we use to render tiles
  */
